from math import floor, ceil

from matplotlib.pyplot import figure, imshow, legend, plot, show
import numpy as np
from numpy import array, flipud, linspace, sqrt
from numpy.linalg import norm
from numpy.random import normal, uniform
from numba import jit, jitclass
from numba import int32, float64
from pathlib import Path
from pickle import dump, load
from scipy.stats import invgauss, gaussian_kde
from sys import argv, exc_info
from time import time
from vis_plume_btc import plot_pickled_kernel, plot_pickled_plume

stem = argv[1]
max_plot_time = 4.0e4
dump_time = np.float64(7e3)

if stem == 'basic' or stem == 'smoothednoctrw':
    simulation_params = {
        'num_particles' : np.int32(40000),
        'inset_frac' : np.float64(0.05) 
    }
else:
    simulation_params = {
        'num_particles' : np.int32(10000),
        'inset_frac' : np.float64(0.05) 
    }

alpha_t = np.float64(0.01)
d = np.float64(0.05)

domain_params = {
    'basic' : {
        'advection' : {
            'mode' : "retard",
            'R' : np.float64(1.0)
        },
        'trapping' : {
            'mode' : 'none',
        },
    },
    'smoothednoctrw' : {
        'advection' : {
            'mode' : "retard",
            'R' : np.float64(1.0)
        },
        'trapping' : {
            'mode' : 'none',
        },
    },
    'smoothed' : {
        'advection' : {
            'mode' : "invgauss",
            'alpha_l': np.float64(0.152)
        },
        'trapping' : {
            'mode' : 'none',
        },
    }
}

print(stem)
dp = domain_params[stem]
if stem == 'smoothednoctrw':
    vstem = 'smoothed'
else:
    vstem = stem

pickle_jar_name = 'v_field_' + vstem + '.pickle'

pickle_file = open(pickle_jar_name,'rb')
pickle_dict = load(pickle_file)
pickle_file.close()

axis_x = pickle_dict['xaxis']
axis_y = pickle_dict['yaxis']
v_x = pickle_dict['vX']
v_y = pickle_dict['vY']
width_x = pickle_dict['width_x']
width_y = pickle_dict['width_y']
dx = axis_x[1] - axis_x[0]
dy = axis_y[1] - axis_y[0]
max_x = axis_x[-1] + dx/2
max_y = axis_y[-1] + dy/2

spec = [
    ('pos', float64[:]),
    ('t', float64)
]
@jitclass(spec)
class Particle(object):
    def __init__(self, init_x, init_y):
        self.pos = np.array([init_x,init_y], dtype=np.float64)
        self.t = np.float64(0)

@jit(float64[:](int32,int32), nopython=True)
def velocity_idx(idx_x, idx_y):
    vel = array([v_x[idx_y, idx_x], v_y[idx_y, idx_x]], dtype=np.float64)
    return vel

@jit(float64[:](float64[:]), nopython=True)
def velocity(pos):
    idx_x = np.int32(floor(pos[0]/max_x*len(axis_x)))
    idx_y = np.int32(floor(pos[1]/max_y*len(axis_y)))

    vel = velocity_idx(idx_x, idx_y)
    return vel

@jit(float64(float64[:]), nopython=True)
def speed(pos):
    sp = norm(velocity(pos))
    return sp

@jit(float64[:](float64[:]), nopython=True)
def stepvector(pos):
    vel = velocity(pos)

    ortho = np.array([-vel[1],vel[0]], dtype=np.float64)
    ortho = ortho/norm(ortho)

    delta_pos = vel/norm(vel)*d + sqrt(2*alpha_t*d)*normal()*ortho
    return delta_pos

if dp['advection']['mode'] == 'invgauss':
    @jit(float64(float64[:]),forceobj=True)
    def advection_time(pos):
        R = invgauss.rvs(2*dp['advection']['alpha_l']/d)
        s = speed(pos)
        at = R*d**2/(2*dp['advection']['alpha_l']*s)
        return(at)
elif dp['advection']['mode'] == 'retard':
    @jit(float64(float64[:]),forceobj=True)
    def advection_time(pos):
        s = speed(pos)
        at = dp['advection']['R']*d/s
        return(at)
else:
    @jit(float64(float64[:]),forceobj=True)
    def advection_time(pos):
        s = speed(pos)
        at = d/s
        return(at)

@jit(float64(float64), nopython=True)
def immobile_time(at):
    return np.float64(0)

@jit(float64(float64[:]),forceobj=True)
def total_time(pos):
    at = advection_time(pos)
    it = immobile_time(at)
    return(at + it)

@jit(forceobj=True)
def step(particle, plume_pos):
    pos = particle.pos

    particle.pos = pos + stepvector(pos)
    delta = total_time(pos)
    if particle.t < dump_time and particle.t + delta > dump_time:
        plume_pos.append(pos)
    particle.t = particle.t + delta

@jit(forceobj=True)
def breakthrough_time(particle, plume_pos):
    while particle.pos[1] > axis_y[1]:
        step(particle, plume_pos)
    return particle.t

def main():
    start_time = time()
    particle_list = []
    breakthrough_times = []
    plume_pos = []

    #Initialize particle list
    assert simulation_params['inset_frac'] < 0.5
    lb_idx_x = floor(simulation_params['inset_frac']*len(axis_x))
    ub_idx_x = ceil((1-simulation_params['inset_frac']) * len(axis_x))-1
    idx_y = len(axis_y)-3
    tmp = []
    for idx_x in range(lb_idx_x,ub_idx_x):
        tmp.append(norm(velocity_idx(idx_x,idx_y)))
        #tmp.append(1.0)
    num_particles = np.ceil(simulation_params['num_particles'] * array(tmp)/sum(tmp))

    for inc_x in range(ub_idx_x-lb_idx_x):
        for count in range(int(num_particles[inc_x])):
            idx_x = lb_idx_x + inc_x
            x = axis_x[idx_x] + dx*uniform(low=-0.5, high=0.5)
            y = axis_y[idx_y] + dy*uniform(low=-0.5, high=0.5)
            particle_list.append(Particle(x,y))

    number = 1
    for particle in particle_list:
        if number % 500 == 1:
            print("PARTICLE", number)
        try:
            btt = breakthrough_time(particle, plume_pos)
            if not (np.isinf(btt) or np.isnan(btt)):
                if not btt > max_plot_time:
                    breakthrough_times.append(btt)
            number += 1
        except KeyboardInterrupt:
            print('\n')
            break
        except:
             e = exc_info()[0]
             print("Error: %s" % e )
             
    print("Traced", number - 1, "particles in", time() - start_time, "seconds")

    #BTC plotting
    breakthrough_times.sort()

    print (np.mean(breakthrough_times))
    kernel = gaussian_kde(breakthrough_times)

    pickle_jar_name = 'btc_' + stem + '.pickle'
    pickle_jar = open(pickle_jar_name, 'wb')
    dump(kernel, pickle_jar)
    pickle_jar.close()

    figure()
    plot_pickled_kernel('basic', max_plot_time)
    plot_pickled_kernel('smoothed', max_plot_time)
    legend()

    #PLUME plotting
    plume_data = np.transpose(np.array(plume_pos))
    plume_kernel = gaussian_kde(plume_data)

    pickle_jar_name = 'plume_' + stem + '.pickle'
    pickle_jar = open(pickle_jar_name, 'wb')
    pickle_dict = {
        'plume_kernel' : plume_kernel,
        'width_x': width_x,
        'width_y': width_y}
    dump(pickle_dict, pickle_jar)
    pickle_jar.close()

    figure()
    plot_pickled_plume(stem)
    show()

main()
