from matplotlib.pyplot import figure, gca, imshow, legend, plot, show, xlabel, ylabel
from matplotlib.ticker import FuncFormatter
from matplotlib.font_manager import FontProperties
from numpy import absolute, amax, array, flipud, format_float_scientific, linspace, meshgrid, reshape, vstack
from numpy.linalg import norm
from pathlib import Path
from pickle import load
from sys import argv

def plot_pickled_kernel(stem, max_plot_time, line_style='solid', fancy_title=None):
    pickle_jar_name =  'btc_' + stem + '.pickle'
    candidate = Path(pickle_jar_name)
    if candidate.exists():
        axis_t = linspace(0, max_plot_time, num=500)
        pickle_file = open(pickle_jar_name,'rb')
        kernel = load(pickle_file)

        btc_vals = kernel(axis_t)

        if fancy_title is None:
            fancy_title = stem
        plot(axis_t, btc_vals, ls=line_style, label=fancy_title)
        xlabel('Time [s]')
        ylabel('Relative concentration')
        ax = gca()

        def formatter(x,pos):
            if x**2 < 1e-10:
                return "0"
            else:
                y = x * (1 + 1.08e-12)
                return format_float_scientific(y, precision=0, exp_digits=1)

        ax.xaxis.set_major_formatter(FuncFormatter(formatter))
        ax.yaxis.set_major_formatter(FuncFormatter(formatter))

def plot_pickled_plume(stem):
    pickle_jar_name =  'plume_' + stem + '.pickle'
    candidate = Path(pickle_jar_name)
    if candidate.exists():
        pickle_file = open(pickle_jar_name,'rb')
        pickle_dict = load(pickle_file)

        plume_kernel = pickle_dict['plume_kernel']
        width_x = pickle_dict['width_x']
        width_y = pickle_dict['width_y']

        axis_x = linspace(0, width_x, 100)
        axis_y = linspace(0, width_y, 100)

        X, Y = meshgrid(axis_x, axis_y)
        positions = vstack([X.ravel(), Y.ravel()])
        Z = reshape(plume_kernel(positions).T, X.shape)
        imshow(flipud(Z), extent=[0, width_x, 0, width_y])
        xlabel('x [m]')
        ylabel('y [m]')

def print_divergence(stem_1, stem_2):
    pickle_jar_1_name =  'plume_' + stem_1 + '.pickle'
    pickle_jar_2_name =  'plume_' + stem_2 + '.pickle'
    if Path(pickle_jar_1_name).exists() and Path(pickle_jar_2_name).exists() :
        #Load first plume
        pickle_file = open(pickle_jar_1_name,'rb')
        pickle_dict = load(pickle_file)

        plume_kernel = pickle_dict['plume_kernel']
        width_x = pickle_dict['width_x']
        width_y = pickle_dict['width_y']

        axis_x = linspace(0, width_x, 100)
        axis_y = linspace(0, width_y, 100)

        X, Y = meshgrid(axis_x, axis_y)
        positions = vstack([X.ravel(), Y.ravel()])

        Z_1 = reshape(plume_kernel(positions).T, X.shape)
        pickle_file.close()

        #Load second plume
        pickle_file = open(pickle_jar_2_name,'rb')
        pickle_dict = load(pickle_file)

        plume_kernel = pickle_dict['plume_kernel']
        
        Z_2 = reshape(plume_kernel(positions).T, X.shape)
        pickle_file.close()
        print("DIVERGENCE OF", stem_1, "VS.", stem_2)
        print ("L2:", str(norm(Z_1 - Z_2)))
        print ("MAX:", str(amax(absolute(Z_1 - Z_2))),"\n")


if __name__ == '__main__':
    max_plot_time = float(argv[1])
    figure()
    plot_pickled_kernel('basic', max_plot_time, fancy_title='Classical physics, highly resolved velocity field')
    plot_pickled_kernel('smoothednoctrw', max_plot_time, line_style=':', fancy_title='Classical physics, smoothed velocity field')
    plot_pickled_kernel('smoothed', max_plot_time, fancy_title='Additional CTRW physics, smoothed velocity field')
    font_props = FontProperties()
    font_props.set_size('small')
    legend(prop=font_props)

    figure()
    plot_pickled_plume('basic')

    figure()
    plot_pickled_plume('smoothednoctrw')

    figure()
    plot_pickled_plume('smoothed')

    show()
    print_divergence('basic','smoothed')
    print_divergence('basic','smoothednoctrw')